package persistent.counter.impl;

import java.util.Calendar;
import java.util.Properties;
import java.util.StringTokenizer;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import javax.xml.namespace.QName;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import org.globus.wsrf.InvalidResourceKeyException;
import org.globus.wsrf.NoSuchResourceException;
import org.globus.wsrf.PersistentResource;
import org.globus.wsrf.Resource;
import org.globus.wsrf.ResourceException;
import org.globus.wsrf.ResourceIdentifier;
import org.globus.wsrf.ResourceLifetime;
import org.globus.wsrf.ResourceKey;
import org.globus.wsrf.ResourceProperties;
import org.globus.wsrf.ResourceProperty;
import org.globus.wsrf.ResourcePropertySet;
import org.globus.wsrf.WSRFConstants;
import org.globus.wsrf.impl.ReflectionResourceProperty;
import org.globus.wsrf.impl.SimpleResourcePropertySet;
import org.globus.wsrf.impl.SimpleResourceProperty;

import persistent.counter.LastIncrementType;

public class CounterResource implements Resource,
                                        ResourceProperties,
                                        ResourceLifetime,
                                        PersistentResource {
    /** The counter service WSDL namespace */
    public static final String COUNTER_NS =
        "http://course.example/Counter";
    /** The qualified name of the counter resource property document */
    public static final QName COUNTER_RP_DOC = new QName(
        COUNTER_NS, "CounterRPDoc"
    );
    /** The qualified name of the value resource property */    
    public static final QName VALUE_RP = new QName(
        COUNTER_NS, "ValueRP"
    );
    /** The qualified name of the History resource property */    
    public static final QName HISTORY_RP = new QName(
        COUNTER_NS, "HistoryRP"
    );
    /** The qualified name of the LastIncrement resource property */    
    public static final QName LAST_INCREMENT_RP = new QName(
        COUNTER_NS, "LastIncrementRP"
    );        
    
    private static Log logger = LogFactory.getLog(
        CounterResource.class.getName()
    );

    private ResourcePropertySet rpDoc = null;
    private ResourceProperty valueRP = null;
    private ResourceProperty historyRP = null;
    private ResourceProperty lastIncrementRP = null;
    private ResourceProperty terminationTimeRP = null;
    private ResourceProperty currentTimeRP = null;            
    private Calendar terminationTime = null;
    private Integer resourceId = null;
    
    public void increment(int incr) throws ResourceException {
	logger.debug("Adding " + incr + " ... ");

        // Update value resource property
        Integer value = (Integer) this.valueRP.get(0);
        this.valueRP.set(0, value.intValue() + incr);

        // Update history resource property
        this.historyRP.add(new Integer(incr));

        // Update LastIncrement resource property
        this.lastIncrementRP.set(
            0, new LastIncrementType(incr, Calendar.getInstance())
        );

        store();
    }

    public void initialize() throws Exception {
        // Create the resource property document for this resource
        this.rpDoc = new SimpleResourcePropertySet(COUNTER_RP_DOC);

        // Create and initialize the value resource property
        this.valueRP = new SimpleResourceProperty(VALUE_RP);
        this.valueRP.add(new Integer(0));

        // Create the history resource property        
        this.historyRP = new SimpleResourceProperty(HISTORY_RP);        

        // Create and initialize the LastIncrement resource property        
        this.lastIncrementRP = new SimpleResourceProperty(LAST_INCREMENT_RP);
        this.lastIncrementRP.add(
            new LastIncrementType(0, Calendar.getInstance())
        );

        // Create resource properties for ScheduledResourceTermination porttype
        this.terminationTimeRP = new ReflectionResourceProperty(
            WSRFConstants.TERMINATION_TIME, "terminationTime", this
        );
        this.currentTimeRP = new ReflectionResourceProperty(
            WSRFConstants.CURRENT_TIME, "currentTime", this
        );
        
        // Add resource properties to resource property document
        this.rpDoc.add(this.valueRP);
        this.rpDoc.add(this.historyRP);
        this.rpDoc.add(this.lastIncrementRP);
        this.rpDoc.add(this.terminationTimeRP);
        this.rpDoc.add(this.currentTimeRP);

        this.resourceId = new Integer(this.hashCode());
        store();
    }   

    public ResourcePropertySet getResourcePropertySet() {
        return this.rpDoc;
    }

    public void setTerminationTime(Calendar time) {
        this.terminationTime = time;
        try {
            store();
        } catch (ResourceException e) {
            logger.error("Failed to store resource: " + e);
        }
    }

    public Calendar getTerminationTime() {
        return this.terminationTime;
    }

    public Calendar getCurrentTime() {
        return Calendar.getInstance();
    }

    public Object getID() {
        return this.resourceId;
    }

    public void load(ResourceKey key) throws ResourceException,
                                             NoSuchResourceException,
                                             InvalidResourceKeyException {
        try {
            this.resourceId = (Integer) key.getValue();
            logger.debug("Trying to load counter: '" + this.resourceId + "'");
            Properties propertyFile = new Properties();
            File file = new File("counter." + resourceId);
            if (!file.exists()) {
                throw new NoSuchResourceException(
                    "Counter " + resourceId + " does not exist"
                );
            }

            propertyFile.load(
                new FileInputStream("counter." + resourceId)
            );

            this.rpDoc = new SimpleResourcePropertySet(COUNTER_RP_DOC);        

            // Load counter value
            Integer value = new Integer(propertyFile.getProperty("counter.value"));
            this.valueRP = new SimpleResourceProperty(VALUE_RP);
            this.valueRP.add(value);        
            
            // Load history values
            this.historyRP = new SimpleResourceProperty(HISTORY_RP);                
            String history = propertyFile.getProperty("counter.history");
            StringTokenizer tokenizer = new StringTokenizer(history, " ");
            while (tokenizer.hasMoreTokens()) {
                this.historyRP.add(new Integer(tokenizer.nextToken()));
            }
            
            // Load last increment value
            Calendar time = Calendar.getInstance();
            time.setTimeInMillis(
                Long.parseLong(
                    propertyFile.getProperty("counter.last.increment.time")
                )
            );
            int lastIncr = Integer.parseInt(
                propertyFile.getProperty("counter.last.increment.value")
            );
            this.lastIncrementRP = new SimpleResourceProperty(LAST_INCREMENT_RP);
            this.lastIncrementRP.add(new LastIncrementType(lastIncr, time));

            // Load lifetime value
            this.terminationTimeRP = new ReflectionResourceProperty(
                WSRFConstants.TERMINATION_TIME, "terminationTime", this
            );
            this.currentTimeRP = new ReflectionResourceProperty(
                WSRFConstants.CURRENT_TIME, "currentTime", this
            );        
            String lifetime = propertyFile.getProperty(
                "counter.termination.time"
            );
            if (lifetime != null) {
                this.terminationTime = Calendar.getInstance();
                this.terminationTime.setTimeInMillis(Long.parseLong(lifetime));
            }
            
            // Add resource properties to resource property document
            this.rpDoc.add(this.valueRP);
            this.rpDoc.add(this.historyRP);
            this.rpDoc.add(this.lastIncrementRP);
            this.rpDoc.add(this.terminationTimeRP);
            this.rpDoc.add(this.currentTimeRP);
        } catch (Exception e) {
            throw new ResourceException(e);
        }        
    }

    // Store counter resource state to file    
    public void store() throws ResourceException {
        logger.debug("Storing resource state for counter: " + this.resourceId);
        Properties propertyFile = new Properties();
        // Add counter value
        propertyFile.setProperty(
            "counter.value", this.valueRP.get(0).toString()
        );

        // Add history value
        String history = "";
        for(int i = 0; i < historyRP.size(); i++) {
            history += historyRP.get(i).toString() + " ";
        }
        propertyFile.setProperty("counter.history", history);

        // Add last increment value
        LastIncrementType lastIncrement = (LastIncrementType)
            this.lastIncrementRP.get(0);
        propertyFile.setProperty(
            "counter.last.increment.time",
            "" + lastIncrement.getTime().getTimeInMillis()
        );
        propertyFile.setProperty(
            "counter.last.increment.value",
            "" + lastIncrement.getIncrement()
        );

        // Add lifetime (if set)
        if (terminationTime != null) {
            propertyFile.setProperty(
                "counter.termination.time",
                "" + terminationTime.getTimeInMillis()
            );
        }
        
        // store to file
        try {
            propertyFile.store(
                new FileOutputStream(
                    "counter." + this.resourceId
                ),
                "persisted counter state"            
            );
        } catch (Exception e) {
            throw new ResourceException(e);
        }
    }

    // Delete resouce state file
    public void remove() throws ResourceException {
        logger.debug("Removing counter resource: " + this.resourceId);
        File resourceFile = new File("counter." + this.resourceId);
        resourceFile.delete();
    }
}
